config = { 
    "decos" : "flow-cutter",
    "decot" : "1", 
    "knowledge_compiler" : "sharpsat-td", 
    "constrained" : "XD",
    "backdoort" : "30",
    "backdoors" : "fvs",
    "number_cores" : "1"
}