INTERNAL_ALGEBRAIC = "internal_algebraic"
EXTERNAL_ALGEBRAIC = "algebraic_atom"
INTERNAL_PENALTY = "internal_penalty"
EXTERNAL_PENALTY = "penalty_atom"

conversions = {
    INTERNAL_ALGEBRAIC : EXTERNAL_ALGEBRAIC,
    INTERNAL_PENALTY : EXTERNAL_PENALTY
}