"""The module aspmc.graph contains all the graph related classes:

    * a hypergraph class in aspmc.graph.hypergraph, which can be used to generated tree decompositions.
    * a binary tree class in aspmc.graph.bintree, which is used for D/Vtrees.
    * a tree decomposition class in aspmc.graph.treedecomposition, which can generated tree decompositions from different inputs.
        Also it offers many different utilities.  
"""

from . import *